/**
 * Copyright (c) 2006-2012 EBM WebSourcing, 2012-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.objectweb.petals.tools.rmi.server.remote.implementations;

import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.TimeUnit;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.MessageExchangeFactory;
import javax.jbi.messaging.MessagingException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.objectweb.petals.tools.rmi.remote.RemoteDeliveryChannel;
import org.objectweb.petals.tools.rmi.remote.RemoteMessageExchangeFactory;
import org.objectweb.petals.tools.rmi.server.remote.implementations.RemoteComponentContextClientFactoryImpl.SharedComponentData;
import org.objectweb.petals.tools.rmi.server.util.Convert;
import org.ow2.petals.commons.log.FlowAttributesExchangeHelper;
import org.ow2.petals.commons.log.PetalsExecutionContext;
import org.ow2.petals.jbi.messaging.exchange.MessageExchangeWrapper;
import org.ow2.petals.jbi.messaging.exchange.PetalsMessageExchange;
import org.ow2.petals.jbi.servicedesc.endpoint.Location;

/**
 * The rmi implementation of delivery channel.
 * 
 * @author Nicolas Salatge - EBM WebSourcing
 */
public class RemoteDeliveryChannelImpl extends UnicastRemoteObject implements RemoteDeliveryChannel {

    /**
     * the serialVersionUID.
     */
    private static final long serialVersionUID = -864958310638895099L;

    /**
     * internal error message.
     */
    private static final String INTERNAL_ERROR_MSG = "This message exchange is not present in the HashMap\n"
            + "Either the message exchange has been " + "created by another message exchange factory\n"
            + "or the message exchange is terminated " + "(It has already been deleted of the HashMap)";

    /**
     * The RMI client id in the JBI environment to associate the message exchanges to the right RMI client of the SE
     * RMI.
     */
    private final String rmiClientId;

    private final SharedComponentData sharedData;

    private final Location location;

    public RemoteDeliveryChannelImpl(final SharedComponentData sharedData, final String rmiClientId)
            throws RemoteException {
        this.sharedData = sharedData;
        this.rmiClientId = rmiClientId;
        this.location = new Location(sharedData.componentContext.getComponentName(), rmiClientId);
    }

    @Override
    public MessageExchange accept() throws MessagingException, RemoteException {
        return this.accept(0);
    }

    @Override
    public MessageExchange accept(final long timeoutMS) throws MessagingException, RemoteException {

        MessageExchange fake = null;

        // realize the accept
        MessageExchange concrete = null;
        try {
            this.sharedData.threadReceivingMessages.add(Thread.currentThread());

            final BlockingQueue<MessageExchange> receivedMessages = this.sharedData.allReceivedMessages
                    .get(this.rmiClientId);
            if (receivedMessages == null) {
                throw new RemoteException(String.format(
                        "Unable to retrieve the queue accepting message exchanges of RMI client '%s'.",
                        this.rmiClientId));
            }

            this.sharedData.logger.fine("Waiting to receive a message for RMI client '" + this.rmiClientId + "' ...");
            if (timeoutMS > 0) {
                concrete = receivedMessages.poll(timeoutMS, TimeUnit.MILLISECONDS);
            } else {
                concrete = receivedMessages.take();
            }
            this.sharedData.logger.fine(String.format("Receive a message '%s' for RMI client '%s'.",
                    concrete != null ? concrete.getExchangeId() : "<null>",
                    this.rmiClientId));

        } catch (final InterruptedException e) {
            throw new RemoteException(e.getMessage(), e);
        } finally {
            this.sharedData.threadReceivingMessages.remove(Thread.currentThread());
        }

        if (concrete != null) {
            // Convert concrete exchange to fake exchange
            try {
                fake = Convert.createFakeMessageExchangeFromConcreteMessageExchange(concrete, this.location);
            } catch (final MessagingException e) {
                throw Convert.convertMessagingException(e);
            }

            // Store the message
            this.storeIfActive(concrete);

            // delete the concrete message of the hashmap if not active
            this.getConcreteMessageOrRemoveIfNotActive(fake);

            this.sharedData.logger.finest("Return the Accept(timeoutMS) method, exchange Id : " + fake.getExchangeId());
        }

        return fake;
    }

    @Override
    public void close() throws MessagingException, RemoteException {
        throw new UnsupportedOperationException("Not implemented.");
    }

    @Override
    public RemoteMessageExchangeFactory createExchangeFactory() throws RemoteException {
        // create the concrete MessageExchangeFactory
        final MessageExchangeFactory mef = this.sharedData.deliveryChannel.createExchangeFactory();

        // Return the fake Message Exchange Factory
        return new RemoteMessageExchangeFactoryImpl(this.sharedData, mef, this.rmiClientId);
    }

    @Override
    public RemoteMessageExchangeFactory createExchangeFactory(final QName interfaceName) throws RemoteException {
        // create the concrete MessageExchangeFactory
        final MessageExchangeFactory mef = this.sharedData.deliveryChannel.createExchangeFactory(interfaceName);

        // Return the fake Message Exchange Factory
        return new RemoteMessageExchangeFactoryImpl(this.sharedData, mef, this.rmiClientId);
    }

    @Override
    public RemoteMessageExchangeFactory createExchangeFactory(final ServiceEndpoint endpoint) throws RemoteException {
        // create the concrete MessageExchangeFactory
        final MessageExchangeFactory mef = this.sharedData.deliveryChannel.createExchangeFactory(endpoint);

        // Return the fake Message Exchange Factory
        return new RemoteMessageExchangeFactoryImpl(this.sharedData, mef, this.rmiClientId);
    }

    @Override
    public RemoteMessageExchangeFactory createExchangeFactoryForService(final QName serviceName) throws RemoteException {
        // create the concrete MessageExchangeFactory
        MessageExchangeFactory mef = this.sharedData.deliveryChannel.createExchangeFactory(serviceName);

        // Return the fake Message Exchange Factory
        return new RemoteMessageExchangeFactoryImpl(this.sharedData, mef, this.rmiClientId);
    }

    @Override
    public void send(final MessageExchange exchange) throws MessagingException, RemoteException {
        // Verification
        if (exchange == null) {
            throw new MessagingException("send: The MessageExchange is null");
        }

        // they should have been set by the RMIClient
        PetalsExecutionContext.putFlowAttributes(FlowAttributesExchangeHelper.getFlowAttributes(exchange));

        if (!(exchange instanceof MessageExchangeWrapper)) {
            throw new MessagingException("send: Invalid object");
        }

        final MessageExchangeWrapper wrapper = (MessageExchangeWrapper) exchange;

        // Get the concrete message and delete of the hasmap
        MessageExchange concrete = this.getConcreteMessageOrRemoveIfNotActive(exchange);

        if (concrete == null) {
            throw new MessagingException(RemoteDeliveryChannelImpl.INTERNAL_ERROR_MSG);
        }

        try {
            // Set all value
            Convert.setFakeValueMEToConcreteValueME(this.sharedData.componentContext, exchange, concrete);

            // Send the concrete exchange
            wrapper.removeOwnership();

            this.sharedData.deliveryChannel.send(concrete);
        } catch (MessagingException e) {
            // Delete the message exchange of the hashmap
            this.sharedData.messages.remove(exchange.getExchangeId());
            throw Convert.convertMessagingException(e);
        } catch (RuntimeException e) {
            // Delete the message exchange of the hashmap
            this.sharedData.messages.remove(exchange.getExchangeId());
            throw Convert.convertRuntimeException(e);
        }

        // Get the concrete message and delete of the hashmap if not active
        this.getConcreteMessageOrRemoveIfNotActive(exchange);

        this.sharedData.logger.finest("Return the send() method, exchange Id : " + exchange.getExchangeId());
    }

    @Override
    public MessageExchange sendSync(final MessageExchange exchange) throws MessagingException, RemoteException {
        return sendSync(exchange, 0);
    }

    @Override
    public MessageExchange sendSync(final MessageExchange exchange, final long timeoutMS)
            throws MessagingException, RemoteException {
        // Verification
        if (exchange == null) {
            throw new MessagingException("sendSync with timeout: The MessageExchange is null");
        }

        // they should have been set by the RMIClient
        PetalsExecutionContext.putFlowAttributes(FlowAttributesExchangeHelper.getFlowAttributes(exchange));

        if (!(exchange instanceof MessageExchangeWrapper)) {
            throw new MessagingException("sendSync with timeout: Invalid object");
        }

        final MessageExchangeWrapper wrapper = (MessageExchangeWrapper) exchange;

        MessageExchange result = null;

        // Get the concrete message and delete of the hasmap
        MessageExchange concrete = this.getConcreteMessageOrRemoveIfNotActive(exchange);

        if (concrete == null) {
            throw new MessagingException(RemoteDeliveryChannelImpl.INTERNAL_ERROR_MSG);
        }

        try {
            // Set all value
            Convert.setFakeValueMEToConcreteValueME(this.sharedData.componentContext, exchange, concrete);

            // Send the concrete exchange
            final PetalsMessageExchange fake = wrapper.removeOwnership();

            if (this.sharedData.deliveryChannel.sendSync(concrete, timeoutMS)) {
                wrapper.returnOwnership(fake);

                // Set all value
                Convert.setConcreteValueMEToFakeValueME(concrete, exchange);

                result = exchange;
            } else {
                result = null;
            }
        } catch (MessagingException e) {
            // Delete the message exchange of the hashmap
            this.sharedData.messages.remove(exchange.getExchangeId());
            throw Convert.convertMessagingException(e);
        } catch (RuntimeException e) {
            // Delete the message exchange of the hashmap
            this.sharedData.messages.remove(exchange.getExchangeId());
            throw Convert.convertRuntimeException(e);
        }

        // Delete the concrete of the hashmap if needed
        this.getConcreteMessageOrRemoveIfNotActive(exchange);

        this.sharedData.logger
                .finest("Return the sendSync(timeoutMS) method, exchange Id : " + exchange.getExchangeId());

        return result;
    }

    /**
     * Check and storage the concrete message exchange if it is not active.
     */
    private void storeIfActive(final MessageExchange concrete) {
        if (concrete.getStatus() == ExchangeStatus.ACTIVE) {
            this.sharedData.messages.put(concrete.getExchangeId(), concrete);
        }
    }

    /*
     * Added methods
     */

    /**
     * Get the concrete message and remove it of the hashmap. if it not active
     * 
     * @return the concrete message exchange or null
     */
    private MessageExchange getConcreteMessageOrRemoveIfNotActive(final MessageExchange fake) {
        MessageExchange concrete = null;
        if (fake.getStatus() != ExchangeStatus.ACTIVE) {
            // The message is not active. It is deleted of the hashmap
            concrete = this.sharedData.messages.remove(fake.getExchangeId());
        } else {
            // The message is active. It is not deleted of the hashmap
            concrete = this.sharedData.messages.get(fake.getExchangeId());
        }
        return concrete;
    }

}
